unit QuadtreeClass;
//*******************************************************
// Quadtree class.
//*******************************************************

interface

uses Windows, Graphics, sysutils, Classes, contnrs;

type
  TNorthOrSouth = (North, South);
  TEastOrWest = (East, West);

  TQPoint = record
    X: double;
    Y: double;
    Data : TList;
  end;
  PQPoint = ^TQPoint;

  TPointerArray = array of Pointer;

  TQuadPoint =  record
    X: double;
    Y: double;
    Distance : double;
    Data : TPointerArray;
  end;
//  TQuadPoint records contain the X and Y coordinates of a location in a
//  TRbwQuadTree, the distance from that point to a location for which a
//  search was performed, and an array of pointers to the data associated
//  with the location.

  TQuadPointArray = array of TQuadPoint;

  TQuadPointInRegion =  record
    X: double;
    Y: double;
    Data : TPointerArray;
  end;
//  TQuadPointInRegion records contain the X and Y coordinates of a location
//  in a TRbwQuadTree, and an array of pointers to the data associated
//  with the location. These points are all located inside the extent of a
//  T2DBlock (see below) that was used in a search.

  TQuadPointInRegionArray = array of TQuadPointInRegion;

  TExpandDirection = (edNorth, edSouth, edEast, edWest);

  T2DBlock = record
    XMin : double;
    XMax : double;
    YMin : double;
    YMax : double;
  end;
//  T2DBlock delineates a region of space. T2DBlock's are used to perform
//  a search for locations inside a TRbwQuadTree.

  EQTreeError = Class(Exception);
//  EQTreeError is a class for exceptions raised by TRbwQuadTree.

  TRbwQuadTree = class;

  TQtreeNode = class(TObject)
  // TQtreeNode is the most of the real work of a TRbwQuadTree.
  private
    FNumPts: Integer;
    procedure ResetCount;
    procedure SetNumPts(const Value: Integer);
    function GetPoints(Index: Integer): TQPoint;
    procedure FindNearestPoints(const CenterX, CenterY: double;
      const Count: Integer; const List: TObjectList);
  protected
    Children               : array [TNorthOrSouth, TEastOrWest] of TQtreeNode;
    Xmin, Xmax, Ymin, Ymax : double;
    Pts                    : array of TQPoint;
    Parent : TQtreeNode;
    Tree : TRbwQuadTree;
    property NumPts  : Integer read FNumPts write SetNumPts;
    procedure AddPoint(X, Y : double; Data : pointer);
    procedure Clear;
    procedure CheckNearbyLeaves(exclude : TQtreeNode;
      var best_leaf : TQtreeNode; X, Y : double; var best_i : integer;
      var best_dist2, best_dist : double);
    constructor Create(ATree: TRbwQuadTree); overload;
    constructor Create(const x_min, x_max, y_min, y_max : double;
      const ParentNode : TQtreeNode; ATree : TRbwQuadTree); overload;
    procedure ExpandBounds(XY : double; ExpandDirection : TExpandDirection);
    function FindClosestPoint(const X, Y: double): TQPoint;
    // FindClosestPoint returns the X,Y, locations and first piece of data in the
    // point closest to the value specified for X and Y.
    procedure FindPoint(var X, Y : double; var Data : pointer);
    procedure FindClosestPointsData(var X, Y : double; var Data : TPointerArray);
    function LocateLeaf(X, Y : double; Siblings : TStack) : TQtreeNode;
    procedure NearPointInLeaf(X, Y : double; var best_i : integer;
      var best_dist2 : double);
    function Xmid : double;
    function Ymid : double;
    property Points[Index: Integer]: TQPoint read GetPoints;
    procedure RemovePoint(X, Y: double; Data: pointer);
    procedure FindPointsInCircle(Const CenterX, CenterY, Radius,
      RadiusSquared : double; const List : TList);
    procedure FindPointsInBlock(const Block: T2DBlock; const List: TList);
  public
    destructor Destroy; override;
  end;

  TRbwQuadTree = class(TComponent)
  private
    FQTreeNode : TQtreeNode;
    FMaxPoints: integer;
    function GetXMax : double;
    function GetXMin : double;
    function GetYMax : double;
    function GetYMin : double;
    procedure SetXMax(const AValue : double);
    procedure SetXMin(const AValue : double);
    procedure SetYMax(const AValue : double);
    procedure SetYMin(const AValue : double);
    function GetCount : integer;
    function GetPoints(Index: Integer): TQuadPoint;
    procedure SetMaxPoints(const Value: integer);
  public
    procedure AddPoint(X, Y : double; Data : pointer);
    // AddPoint stores Data at location X, Y.
    //
    // If there is already a point at that location the number of pointers
    // stored at that point will be increased by one and the new data will be
    // added at the end of the list of data for the point. Count will be
    // unaffected.
    //
    // If there is not already a point at that location, one will be created
    // and the data will be associated with it, and the Count will be
    // incremented by one.
    //
    // If required, Xmax, XMin, Ymax, or Ymin will be adjusted to accomodate the
    // new data point.
    //
    // The Data associated with a location is not owned by the TRbwQuadTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
    Procedure Clear;
    // Clear deletes all data locations in the TRbwQuadTree and resets the
    // Count to 0.
    //
    // The Data associated with a location is not owned by the TRbwQuadTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
    property Count : integer read GetCount;
    // Count is the number of unique locations stored in a TRbwQuadTree.
    constructor Create(AOwner: TComponent); override;
    // Creates an instance of TRbwQuadTree;
    destructor Destroy; override;
    // Destroys and instance of TRbwQuadTree. Applications should not call
    // Destroy directly. Instead call Free;
    procedure FindClosestPointsData(var X, Y : double;
      var Data : TPointerArray);
    // FindClosestPointsData locates the location nearest to the X, Y values
    // provided as input.  It than returns the X and Y values of that location.
    // The length of Data is changed to the number of pointers stored at
    // the location that was found and the Data pointers are copied into Data.
    procedure FindNearestPoints(const CenterX, CenterY: double;
      const Count: Integer; var Points : TQuadPointArray);
    // FindNearestPoints finds the Count locations in the TRbwQuadTree that are
    // closest to CenterX, CenterY and returns them and their associated data
    // in Points. In the event of ties, the length of Points may be larger than
    // Count.  If the number of locations in the TRbwQuadTree is less than
    // Count, all the locations will be returned.  Points is sorted from the
    // closest locations to CenterX, CenterY to those that are most distant.
    //
    // See also FirstNearestPoint, NearestPointsFirstData
    procedure FindPointsInBlock(const Block: T2DBlock;
      var Points: TQuadPointInRegionArray);
    // FindPointsInBlock finds all locations stored in the TRbwQuadTree that
    // are inside Block and returns them in Points along with their associated
    // data. Points is not sorted.
    procedure FindPointsInCircle(Const CenterX, CenterY, Radius : double;
      var Points : TQuadPointInRegionArray);
    // FindPointsInCircle finds all the locations stored in the TRbwQuadTree
    // that are within a distance of Radius from CenterX, CenterY and returns
    // them and their associated data in Points.  Points is not sorted.
    procedure FirstNearestPoint(var X, Y : double; var Data : pointer);
    // FirstNearestPoint, finds the location in the TRbwQuadTree that is
    // closest to X, Y and returns it in X, Y.  The associated data is returned
    // in Data. In the event of ties in location, the location that is returned
    // is chosen arbitrarily from the locations that are closest. If more than
    // one piece of data is stored at X, Y, the one that is returned is chosen
    // arbitrarily.
    //
    // See also: FindNearestPoints, NearestPointsFirstData
    function NearestPointsFirstData(X, Y : double) : pointer;
    // NearestPointsFirstData finds the location stored in the TRbwQuadTree
    // that is closest to X, Y and returns the first piece of data associated
    // with that location.
    //
    // See also: FindNearestPoints, FirstNearestPoint
    property Points[Index: Integer]: TQuadPoint read GetPoints;
    // Points can be treated as an indexed array of the locations stored
    // in the TRbwQuadTree.
    procedure RemovePoint(X, Y : double; Data : pointer);
    // RemovePoint removes Data at location X, Y.
    //
    // If the number of data pointers stored at X, Y is greater than 1, Data
    // will be deleted from the list of data pointers and Count will be
    // unaffected.
    //
    // If there is only one data pointer stored at X, Y, that location will
    // be deleted and the Count will be decremented by one.
    //
    // The Data associated with a location is not owned by the TRbwQuadTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
  published
    property MaxPoints : integer read FMaxPoints write SetMaxPoints;
    // MaxPoints is the maximum number of data locations in any node of a
    // TRbwQuadTree. MaxPoints affects the performance of the TRbwQuadTree
    // in the retrieval of data in a manner that depends on the particular
    // application.
    Property Xmax : double read GetXMax write SetXMax;
    // Xmax is the largest X value of any of the data points that are expected
    // to be or have already been added to the TRbwQuadTree.  Although it is
    // not required that Xmax be greater than the X coordinate of all locations
    // added to the TRbwQuadTree, performance may suffer if too many of them
    // are greater than Xmax.
    Property XMin : double read GetXMin write SetXMin;
    // XMin is the smallest X value of any of the data points that are expected
    // to be or have already been added to the TRbwQuadTree.  Although it is
    // not required that XMin be less than the X coordinate of all locations
    // added to the TRbwQuadTree, performance may suffer if too many of them
    // are less than XMin.
    Property Ymax : double read GetYMax write SetYMax;
    // Ymax is the largest Y value of any of the data points that are expected
    // to be or have already been added to the TRbwQuadTree.  Although it is
    // not required that Ymax be greater than the Y coordinate of all locations
    // added to the TRbwQuadTree, performance may suffer if too many of them
    // are greater than Ymax.
    Property Ymin : double read GetYMin write SetYMin;
    // Ymin is the smallest Y value of any of the data points that are expected
    // to be or have already been added to the TRbwQuadTree.  Although it is
    // not required that Ymin be less than the Y coordinate of all locations
    // added to the TRbwQuadTree, performance may suffer if too many of them
    // are less than Ymin.
  end;

procedure Register;
  
implementation

type
  TSelectNode = class(TObject)
    Distance : double;
    Point : TQPoint;
  end;

function SelectNodeCompare(Item1, Item2: Pointer): Integer;
var
  Node1, Node2 : TSelectNode;
  Delta : double;
begin
  Node1 := Item1;
  Node2 := Item2;
  Delta := Node2.Distance - Node1.Distance;
  if Delta > 0 then
  begin
    result := -1;
  end
  else if Delta < 0 then
  begin
    result := 1;
  end
  else
  begin
    result := 0;
  end;

end;

// Create the node.
constructor TQtreeNode.Create(ATree : TRbwQuadTree);
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
begin
  inherited Create;
  Parent := nil;
  Xmin := 0;
  Xmax := 0;
  Ymin := 0;
  Ymax := 0;
  Tree := ATree;
  for ns := North to South do
  begin
    for ew := East to West do
    begin
      Children[ns, ew] := nil;
    end;

  end;
  FNumPts := 0;

  // Allocate space for the points.
  SetLength(Pts, Tree.MaxPoints);
end;

// Create and initialize the node.
constructor TQtreeNode.Create(const x_min, x_max, y_min, y_max : double;
  const ParentNode : TQtreeNode; ATree : TRbwQuadTree);
begin
  Create(ATree);
  if not ((x_min <= x_max) and (y_min <= y_max)) then
  begin
    raise EQTreeError.Create('Error: attempt to create a TQtreeNode with '
      + 'an invalid range.');
  end;
  Parent := ParentNode;
  Xmin := x_min;
  Xmax := x_max;
  Ymin := y_min;
  Ymax := y_max;
end;

// Free all children.
destructor TQtreeNode.Destroy;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  Index : integer;
//  ListIndex : integer;
begin
  // If we have no children, free the point memory.
  if (Children[North, West] = nil) then
  begin
    // If required, destroy the data.
    for Index := 0 to NumPts-1 do
    begin
      Pts[Index].Data.Free;
    end;
    SetLength(Pts, 0);
  end
  else
  begin
      // Otherwise free the children.
      for ns := North to South do
          for ew := East to West do
              Children[ns, ew].Free;
      // The Point memory doesn't have to be freed because it was
      // freed when the children were created;
  end;

  inherited Destroy;
end;

procedure TQtreeNode.RemovePoint(X, Y : double; Data : pointer);
var
  xmidLocal, ymidLocal : double;
  ns            : TNorthOrSouth;
  ew            : TEastOrWest;
  AQuadNode : TQtreeNode;
  Index, PointIndex : integer;
begin
  // If the point is outside the range of the TQtreeNode, expand the TQtreeNode to encompass it.
  if (X < Xmin) or (X > Xmax) or (Y < Ymin) or (Y > Ymax) then Exit;

  AQuadNode := self;
  // See if the node belongs in a child.
  While (AQuadNode.Children[North, West] <> nil) do

  begin
    xmidLocal := AQuadNode.Xmid;
    ymidLocal := AQuadNode.Ymid;

    // Check that children exist.

    // Add the point to the correct child.
    if (Y <= ymidLocal) then
        ns := South
    else
        ns := North;
    if (X <= xmidLocal) then
        ew := West
    else
        ew := East;
    AQuadNode := AQuadNode.Children[ns, ew];
  end;

  // find point.
  for Index := 0 to AQuadNode.NumPts -1 do
  begin
    if (AQuadNode.Pts[Index].X = X) and (AQuadNode.Pts[Index].Y = Y) then
    begin
      // Remove the data
      AQuadNode.Pts[Index].Data.Remove(Data);
{      if Assigned(Tree.FDestroyProcedure) then
      begin
        Tree.FDestroyProcedure(Tree, Data);
      end;   }
      // if the number of data items associated with the data
      // point has bee reduced to 0, remove the data point.
      if AQuadNode.Pts[Index].Data.Count = 0 then
      begin
        for PointIndex := Index + 1 to AQuadNode.NumPts -1 do
        begin
          AQuadNode.Pts[PointIndex-1] := AQuadNode.Pts[PointIndex];
        end;
        AQuadNode.Pts[Index].Data.Free;
        AQuadNode.Pts[Index].Data := nil;
        AQuadNode.NumPts := AQuadNode.NumPts -1;
      end;
      Exit;
    end;
  end;
end;

procedure TQtreeNode.AddPoint(X, Y : double; Data : pointer);
var
    xmidLocal, ymidLocal : double;
  i : integer;
    ns            : TNorthOrSouth;
    ew            : TEastOrWest;
  AQuadNode, Child : TQtreeNode;
  Index : integer;
  APoint : TQPoint;
begin
    // If the point is outside the range of the TQtreeNode, expand the TQtreeNode to encompass it.
    if X < Xmin then ExpandBounds(X, edWest);
    if X > Xmax then ExpandBounds(X, edEast);
    if Y < Ymin then ExpandBounds(Y, edSouth);
    if Y > Ymax then ExpandBounds(Y, edNorth);
    
    AQuadNode := self;
    // See if the node belongs in a child.
    While (AQuadNode.NumPts >= Tree.MaxPoints) do

    begin
        xmidLocal := AQuadNode.Xmid;
        ymidLocal := AQuadNode.Ymid;

        // See if we need to create new children.
        if (AQuadNode.Children[North, West] = nil) then
        begin
            // Split the node into four children.
            AQuadNode.Children[North, West] := TQtreeNode.Create
              (AQuadNode.Xmin, xmidLocal, ymidLocal, AQuadNode.Ymax,
              AQuadNode, Tree);
            AQuadNode.Children[North, East] := TQtreeNode.Create
              (xmidLocal, AQuadNode.Xmax, ymidLocal, AQuadNode.Ymax,
              AQuadNode, Tree);
            AQuadNode.Children[South, West] := TQtreeNode.Create
              (AQuadNode.Xmin, xmidLocal, AQuadNode.Ymin, ymidLocal,
              AQuadNode, Tree);
            AQuadNode.Children[South, East] := TQtreeNode.Create
              (xmidLocal, AQuadNode.Xmax, AQuadNode.Ymin, ymidLocal,
              AQuadNode, Tree);

            // Move the old points into the new children.
            for i := 0 to AQuadNode.NumPts-1 do
            begin
              APoint := AQuadNode.Pts[i];
              if (APoint.Y <= ymidLocal) then
                  ns := South
              else
                  ns := North;
              if (APoint.X <= xmidLocal) then
                  ew := West
              else
                  ew := East;
              Child := AQuadNode.Children[ns, ew];
              Child.Pts[Child.NumPts] := APoint;
              Child.FNumPts := Child.FNumPts + 1;
            end; // End moving points to the new children.

            // Free point memory.
            SetLength(AQuadNode.Pts, 0);

        end; // End creating new children.

        // Add the point to the correct child.
        if (Y <= ymidLocal) then
            ns := South
        else
            ns := North;
        if (X <= xmidLocal) then
            ew := West
        else
            ew := East;
        AQuadNode := AQuadNode.Children[ns, ew];
    end;

    // Don't add duplicate point. Instead add data to existing point.
    for Index := 0 to AQuadNode.NumPts -1 do
    begin
      if (AQuadNode.Pts[Index].X = X) and (AQuadNode.Pts[Index].Y = Y) then
      begin
        AQuadNode.Pts[Index].Data.Add(Data);
        Exit;
      end;
    end;

    // Place the point in this node.
    AQuadNode.Pts[AQuadNode.NumPts].X    := X;
    AQuadNode.Pts[AQuadNode.NumPts].Y    := Y;
    AQuadNode.Pts[AQuadNode.NumPts].Data := TList.Create;
    AQuadNode.Pts[AQuadNode.NumPts].Data.Add(Data);
    AQuadNode.NumPts := AQuadNode.NumPts + 1;
end;


// Find the point closest to the given coordinates.
function TQtreeNode.FindClosestPoint(const X, Y : double) : TQPoint;
var
  best_dist2, best_dist : double;
  best_i     : Integer;
  leaf, Aleaf                   : TQtreeNode;
  Siblings : TStack;
begin
  // See which leaf contains the point.
  Siblings := TStack.Create;
  try

    leaf := LocateLeaf(X, Y, Siblings);

    // Find the closest point within the leaf.
    leaf.NearPointInLeaf(X, Y, best_i, best_dist2);

    // If no point was found, check the siblings of the leaf.
    if best_i < 0 then
    begin
      while Siblings.Count > 0 do
      begin
        Aleaf := Siblings.Pop;
        leaf := Aleaf.LocateLeaf(X, Y, Siblings);

        // Find the closest point within the leaf.
        leaf.NearPointInLeaf(X, Y, best_i, best_dist2);
        if best_i >= 0 then
        begin
          Break;
        end;
      end;
    end;

    best_dist := Sqrt(best_dist2);
    // Check nearby leaves for closer points.
    CheckNearbyLeaves(leaf, leaf, X, Y, best_i, best_dist2, best_dist);

    Assert((best_i >= 0) and (best_i < Length(leaf.Pts)));
    result    := leaf.Pts[best_i];
  finally
    Siblings.Free;
  end;
end;

procedure TQtreeNode.FindPoint(var X, Y : double; var Data : pointer);
var
  APoint : TQPoint;
begin
  APoint := FindClosestPoint(X,Y);
  X    := APoint.X;
  Y    := APoint.Y;
  if (APoint.Data <> nil) and (APoint.Data.Count > 0) then
  begin
    Data := APoint.Data[0];
  end
  else
  begin
    Data := nil;
  end;
end;

// See what leaf contains the point.
function TQtreeNode.LocateLeaf(X, Y : double; Siblings : TStack) : TQtreeNode;
var
    ns, nsi         : TNorthOrSouth;
    ew, ewi         : TEastOrWest;
begin
  Result := Self;

  // Search the appropriate child.
  while result.Children[North, West] <> nil do
  begin
    if (Y <= result.ymid) then
        ns := South
    else
        ns := North;
    if (X <= result.xmid) then
        ew := West
    else
        ew := East;

    // if the appropriate child doesn't have any points, we may
    // have to seach the child's siblings so save them.
    for nsi := North to South do
    begin
      for ewi := East to West do
      begin
        if (nsi <> ns) or (ewi <> ew) then
        begin
          Siblings.Push(result.Children[nsi, ewi]);
        end;
      end;
    end;


    Result := result.Children[ns, ew];
  end;
end;

// Return the index of the point closest to the given
// coordinates in this leaf.
procedure TQtreeNode.NearPointInLeaf(X, Y : double; var best_i : Integer;
  var best_dist2 : double);
var
  i : Longint;
  dist2 : double;
  dx, dy : double;
begin
  // If there are no points in the node, set best_i to -1 to indicate no
  // points were found.
  best_dist2 := 0;
  best_i := -1;

  if NumPts > 0 then
  begin
    dx := X - Pts[0].X;
    dy := Y - Pts[0].Y;
    best_dist2 := dx * dx + dy * dy;
    best_i := 0;

    for i := 1 to NumPts-1 do
    begin
      dx := X - Pts[i].X;
      dy := Y - Pts[i].Y;
      dist2 := dx * dx + dy * dy;
      if (best_dist2 > dist2) then
      begin
        best_i := i;
        best_dist2 := dist2;
      end;
    end;
  end;
end;

// Check nearby leaves to see if there is a better point.
procedure TQtreeNode.CheckNearbyLeaves(exclude : TQtreeNode;
  var best_leaf : TQtreeNode; X, Y : double; var best_i : integer;
  var best_dist2, best_dist : double);
var
  xmidLocal, ymidLocal : double;
  i : Integer;
  dist2 : double;
begin
  // If we are excluding this leaf, do nothing.
  if (exclude = Self) then exit;

  // If this is a leaf node, look for close nodes.
  if (Children[North, West] = nil) then
  begin
    NearPointInLeaf(X, Y, i, dist2);
    // Use the closest point in this leaf if
    // 1. there is at least one point in this leaf (i > -1) and either
    // 2. the closest point in this leaf is closer to the search point than
    //    the point that was already found (best_dist2 > dist2) or
    // 3. no point has been found yet (best_i < 0).
    if (i > -1) and ((best_dist2 > dist2) or (best_i < 0)) then
    begin
      best_dist2 := dist2;
      best_dist := Sqrt(best_dist2);
      best_leaf := Self;
      best_i := i;
    end;
  end
  else
  begin
    // Examine children that lie within best_dist
    // of the point.
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the SouthWest child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
        Children[South, West].CheckNearbyLeaves(
          exclude, best_leaf, X, Y, best_i,
          best_dist2, best_dist);
    end;

    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the NorthWest child is close enough.
      if (Y + best_dist > ymidLocal) or (best_i < 0) Then
        Children[North, West].CheckNearbyLeaves(
          exclude, best_leaf, X, Y, best_i,
          best_dist2, best_dist);
    end;

    if (best_i < 0) or (X + best_dist > xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the SouthEast child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
        Children[South, East].CheckNearbyLeaves(
          exclude, best_leaf, X, Y, best_i,
          best_dist2, best_dist);
    end;

    if (best_i < 0) or (X + best_dist > xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the NorthEast child is close enough.
      if (Y + best_dist > ymidLocal) or (best_i < 0) Then
        Children[North, East].CheckNearbyLeaves(
          exclude, best_leaf, X, Y, best_i,
          best_dist2, best_dist);
    end;
  end; // End if a leaf ... else check children ...
end;

procedure TQtreeNode.ExpandBounds(XY : double;
  ExpandDirection : TExpandDirection);
begin
  case ExpandDirection of
    edNorth:
      begin
        Ymax := XY;
        if Children[North, East] <> nil then
        begin
          Children[North, East].ExpandBounds(XY, ExpandDirection);
          Children[North, West].ExpandBounds(XY, ExpandDirection);
        end;
      end; 
    edSouth: 
      begin
        Ymin := XY;
        if Children[North, East] <> nil then
        begin
          Children[South, East].ExpandBounds(XY, ExpandDirection);
          Children[South, West].ExpandBounds(XY, ExpandDirection);
        end; 
      end; 
    edEast: 
      begin
        Xmax := XY;
        if Children[North, East] <> nil then
        begin
          Children[North, East].ExpandBounds(XY, ExpandDirection);
          Children[South, East].ExpandBounds(XY, ExpandDirection);
        end; 
      end;
    edWest:
      begin
        Xmin := XY;
        if Children[North, East] <> nil then
        begin
          Children[North, West].ExpandBounds(XY, ExpandDirection);
          Children[South, West].ExpandBounds(XY, ExpandDirection);
        end; 
      end; 
  end; 
end;

function TQtreeNode.Xmid : double;
begin
  if Children[North, West] = nil then
  begin
    result := (XMax + XMin) / 2
  end
  else
  begin
    result := Children[North, West].XMax;
  end;
end;

function TQtreeNode.Ymid : double;
begin
  if Children[North, East] = nil then
  begin
    result := (YMax + YMin) / 2
  end
  else
  begin
    result := Children[North, East].YMin;
  end;
end;

{ TRbwQuadTree }

function TRbwQuadTree.GetXMax : double;
begin
  result := FQTreeNode.XMax;
end;

function TRbwQuadTree.GetXMin : double;
begin
  result := FQTreeNode.XMin;
end;

function TRbwQuadTree.GetYMax : double;
begin
  result := FQTreeNode.YMax;
end;

function TRbwQuadTree.GetYMin : double;
begin
  result := FQTreeNode.YMin;
end;


procedure TRbwQuadTree.SetXMax(const AValue : double);
begin
  if Count = 0 then
  begin
    FQTreeNode.Xmax := AValue;
  end
  else
  begin
    if (AValue < XMax) then
    begin
      raise EQTreeError.Create('Error: The maximum X value can not be decreased; '
        + 'only increased.');
    end;
    FQTreeNode.ExpandBounds(AValue, edWest);
  end;
end;

procedure TRbwQuadTree.SetXMin(const AValue : double);
begin
  if Count = 0 then
  begin
    FQTreeNode.Xmin := AValue;
  end
  else
  begin
    if (AValue > XMin) then
    begin
      raise EQTreeError.Create('Error: The minimum X value can not be increased; '
        + 'only decreased.');
    end;
    FQTreeNode.ExpandBounds(AValue, edEast);
  end;
end;

procedure TRbwQuadTree.SetYMax(const AValue : double);
begin
  if Count = 0 then
  begin
    FQTreeNode.Ymax := AValue;
  end
  else
  begin
    if (AValue < YMax) then
    begin
      raise EQTreeError.Create('Error: The maximum Y value can not be decreased; '
        + 'only increased.');
    end;
    FQTreeNode.ExpandBounds(AValue, edNorth);
  end;
end;

procedure TRbwQuadTree.SetYMin(const AValue : double);
begin
  if Count = 0 then
  begin
    FQTreeNode.Ymin := AValue;
  end
  else
  begin
    if (AValue > YMin) then
    begin
      raise EQTreeError.Create('Error: The minimum Y value can not be increased; '
        + 'only decreased.');
    end;
    FQTreeNode.ExpandBounds(AValue, edSouth);
  end;
end;

procedure TRbwQuadTree.AddPoint(X, Y : double; Data : pointer);
begin
  FQTreeNode.AddPoint(X, Y, Data );
end;

constructor TRbwQuadTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMaxPoints := 100;
//  FDestroyProcedure := nil;
  FQTreeNode := TQTreeNode.Create(self);
end;

destructor TRbwQuadTree.Destroy;
begin
  FQTreeNode.Free;
  inherited;
end;

function TRbwQuadTree.NearestPointsFirstData(X, Y : double) : pointer;
begin
  if FQTreeNode.NumPts > 0 then
  begin
    FirstNearestPoint(X, Y, result);
  end
  else
  begin
    result := nil;
  end;
end;

procedure TRbwQuadTree.FirstNearestPoint(var X, Y : double; var Data : pointer);
begin
  if FQTreeNode.NumPts > 0 then
  begin
    FQTreeNode.FindPoint(X, Y, Data);
  end
  else
  begin
    raise EQTreeError.Create('Error: No data points in QuadTree.');
  end;
end;

procedure TQtreeNode.SetNumPts(const Value: Integer);
var
  OldNumPoints {, Delta} : Integer;
  ew : TEastOrWest;
  ns : TNorthOrSouth;
  AChild : TQtreeNode;
  Index : integer;
  PtCount : integer;
begin
  if FNumPts <> Value then
  begin
    OldNumPoints := FNumPts;
    FNumPts := Value;

    if (FNumPts <= Tree.MaxPoints) and (OldNumPoints > Tree.MaxPoints) then
    begin
      SetLength(Pts, Tree.MaxPoints);
      PtCount := 0;
      for ns := North to South do
      begin
        for ew := East to West do
        begin
          AChild := Children[ns, ew];
          for Index := 0 to AChild.NumPts -1 do
          begin
            Assert(PtCount < Tree.MaxPoints);
            Pts[PtCount] :=  AChild.Pts[Index];
            Inc(PtCount);
          end;
          AChild.FNumPts := 0;
          SetLength(AChild.Pts, 0);
          AChild.Free;
          Children[ns, ew] := nil;
        end;
      end;
      Assert(PtCount = FNumPts);
    end;

    if Parent <> nil then
    begin
      Parent.ResetCount;
    end;
  end;
end;

function TQtreeNode.GetPoints(Index: Integer): TQPoint;
var
  ew : TEastOrWest;
  ns : TNorthOrSouth;
  AChild : TQtreeNode;
begin
  if Children[North,East] = nil then
  begin
    result := Pts[Index];
  end
  else
  begin
    for ew := West downto East do
    begin
      for ns := North to South do
      begin
        AChild := Children[ns,ew];
        if AChild.NumPts > Index then
        begin
          result := AChild.GetPoints(Index);
          Exit;
        end
        else
        begin
          Index := Index - AChild.NumPts;
        end;
      end;
    end;
  end;
end;

procedure TQtreeNode.FindClosestPointsData(var X, Y: double;
  var Data: TPointerArray);
var
  APoint : TQPoint;
  Index : integer;
begin
  APoint := FindClosestPoint(X,Y);
  X    := APoint.X;
  Y    := APoint.Y;
  SetLength(Data, APoint.Data.Count);
  for Index := 0 to APoint.Data.Count -1 do
  begin
    Data[Index] := APoint.Data[Index];
  end;
end;

procedure TRbwQuadTree.FindClosestPointsData(var X, Y: double;
  var Data: TPointerArray);
begin
  if FQTreeNode.NumPts > 0 then
  begin
    FQTreeNode.FindClosestPointsData(X, Y, Data);
  end
  else
  begin
    raise EQTreeError.Create('Error: No data points in QuadTree.');
  end;
end;

function TRbwQuadTree.GetCount: integer;
begin
  result := FQTreeNode.NumPts;
end;

procedure TRbwQuadTree.RemovePoint(X, Y: double; Data: pointer);
begin
  FQTreeNode.RemovePoint(X, Y, Data );
end;

function TRbwQuadTree.GetPoints(Index: Integer): TQuadPoint;
var
  QPoint : TQPoint;
  DataIndex : integer;
begin
  if Index < 0 then
  begin
    raise EQTreeError.Create('Invalid point index < 0.');
  end;
  if Index >= Count then
  begin
    raise EQTreeError.Create('Invalid point index >= Count.');
  end;
  QPoint := FQTreeNode.Points[Index];
  result.X := QPoint.X;
  result.Y := QPoint.Y;
  result.Distance := 0;
  SetLength(result.Data, QPoint.Data.Count);
  for DataIndex := 0 to QPoint.Data.Count -1 do
  begin
    result.Data[DataIndex] := QPoint.Data[DataIndex];
  end;
end;

procedure TQtreeNode.ResetCount;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  Count : integer;
begin
  if Children[North, West] <> nil then
  begin
    Count := 0;
    for ns := North to South do
    begin
      for ew := East to West do
      begin
        Count := Count + Children[ns, ew].NumPts;
      end;
    end;
    NumPts := Count;
  end;
end;

{procedure TRbwQuadTree.SetDestroyProcedure(
  const Value: TDestroyDataProcedure);
begin
  FDestroyProcedure := Value;
end; }

procedure TQtreeNode.FindNearestPoints(const CenterX, CenterY : double;
  const Count: Integer; const List: TObjectList);
var
  Radius : double;
  Node : TSelectNode;
  distance : double;
  APoint : TQPoint;
  dx, dy : double;
  ALeaf : TQtreeNode;
  Siblings : TStack;
  Procedure GetPoints;
  var
    Index, InnerIndex : integer;
  begin
    for Index := 0 to ALeaf.NumPts -1 do
    begin
      APoint := ALeaf.Pts[Index];
      dx := APoint.X - CenterX;
      dy := APoint.Y - CenterY;
      distance := Sqrt(Sqr(dx) + Sqr(dy));
      if (List.Count < Count) or (distance <= Radius) then
      begin
        Node := TSelectNode.Create;
        Node.Distance := distance;
        Node.Point := APoint;
        List.Add(Node);
        if List.Count mod Count = 0 then
        begin
          List.Sort(SelectNodeCompare);
          if List.Count >= Count then
          begin
            Node := List[Count-1] as TSelectNode;
            Radius := Node.Distance;
          end;
          for InnerIndex := Count to List.Count - 1 do
          begin
            Node := List[InnerIndex] as TSelectNode;
            if Node.Distance > Radius then
            begin
              List.Count := InnerIndex;
              break;
            end;
          end;
        end;
      end;
    end;
  end;
begin
  Siblings := TStack.Create;
  try
    ALeaf := LocateLeaf(CenterX, CenterY, Siblings);
    radius := 0;
    GetPoints;
    While Siblings.Count > 0 do
    begin
      ALeaf := Siblings.Pop;

      if (List.Count < Count) or
        ((CenterX + radius >= ALeaf.Xmin) and
        (CenterX - radius <= ALeaf.Xmax) and
        (CenterY + radius >= ALeaf.Ymin) and
        (CenterY - radius <= ALeaf.Ymax)) then
      begin
        ALeaf := ALeaf.LocateLeaf(CenterX, CenterY, Siblings);
        GetPoints;
      end;
    end;
  finally
    Siblings.Free;
  end;
end;

procedure TQtreeNode.FindPointsInCircle(const CenterX, CenterY,
  Radius, RadiusSquared: double; const List: TList);
var
  AChild : TQtreeNode;
  xmidLocal, ymidLocal : double;
  Index : integer;
  dx, dy : double;
  APoint : TQPoint;
  WholeBlockInside : boolean;
begin
  if Children[North, West] = nil then
  begin
    // see if it is worth trying to determine if all the points
    // are inside the circle;
    WholeBlockInside := NumPts > 4;

    // Test each corner to see whether it is inside the circle
    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMax - CenterY;
      if Sqr(dx) + Sqr(dy) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMin - CenterY;
      if Sqr(dx) + Sqr(dy) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMax - CenterY;
      if Sqr(dx) + Sqr(dy) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMin - CenterY;
      if Sqr(dx) + Sqr(dy) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    // If the entire block is inside the circle, add all the points.
    // Otherwise, test each point to see if it is inside the circle and
    // add those points.
    if WholeBlockInside then
    begin
      for Index := 0 to NumPts -1 do
      begin
        List.Add(Addr(Pts[Index]));
      end;
    end
    else
    begin
      for Index := 0 to NumPts -1 do
      begin
        APoint := Pts[Index];
        dx := APoint.X - CenterX;
        dy := APoint.Y - CenterY;
        if Sqr(dx) + Sqr(dy) <= RadiusSquared then
        begin
          List.Add(Addr(Pts[Index]));
        end;
      end;
    end;
  end
  else
  begin
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    if (CenterX - Radius <= xmidLocal) then
    begin
      // West Children are candidates.
      if CenterY - Radius <= ymidLocal then
      begin
        //South Children are candidates.
        AChild := Children[South, West];
        AChild.FindPointsInCircle(CenterX, CenterY, Radius, RadiusSquared, List);
      end;

      if CenterY + Radius >= ymidLocal then
      begin
        //North Children are candidates.
        AChild := Children[North, West];
        AChild.FindPointsInCircle(CenterX, CenterY, Radius, RadiusSquared, List);
      end;
    end;

    if (CenterX + Radius >= xmidLocal) then
    begin
      // East Children are candidates.
      if CenterY - Radius <= ymidLocal then
      begin
        //South Children are candidates.
        AChild := Children[South, East];
        AChild.FindPointsInCircle(CenterX, CenterY, Radius, RadiusSquared, List);
      end;

      if CenterY + Radius >= ymidLocal then
      begin
        //North Children are candidates.
        AChild := Children[North, East];
        AChild.FindPointsInCircle(CenterX, CenterY, Radius, RadiusSquared, List);
      end;
    end
  end;
end;

procedure TQtreeNode.FindPointsInBlock(const Block: T2DBlock;
  const List: TList);
var
  AChild : TQtreeNode;
  xmidLocal, ymidLocal : double;
  Index : integer;
//  dx, dy : double;
  APoint : TQPoint;
begin
  if Children[North, West] = nil then
  begin
    if (Block.XMin <= XMin) and (Block.XMax >= XMax)
      and (Block.YMin <= YMin) and (Block.YMax >= YMax) then
    begin
      for Index := 0 to NumPts -1 do
      begin
        List.Add(Addr(Pts[Index]));
      end;
    end
    else
    begin
      for Index := 0 to NumPts -1 do
      begin
        APoint := Pts[Index];
        if (Block.XMin <= APoint.X) and (Block.XMax >= APoint.X)
          and (Block.YMin <= APoint.Y) and (Block.YMax >= APoint.Y) then
        begin
          List.Add(Addr(Pts[Index]));
        end;
      end;
    end;
  end
  else
  begin
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    if (Block.XMin <= xmidLocal) then
    begin
      // West Children are candidates.
      if Block.YMin <= ymidLocal then
      begin
        //South Children are candidates.
        AChild := Children[South, West];
        AChild.FindPointsInBlock(Block, List);
      end;

      if Block.YMax >= ymidLocal then
      begin
        //North Children are candidates.
        AChild := Children[North, West];
        AChild.FindPointsInBlock(Block, List);
      end;
    end;

    if (Block.XMax >= xmidLocal) then
    begin
      // East Children are candidates.
      if Block.YMin <= ymidLocal then
      begin
        //South Children are candidates.
        AChild := Children[South, East];
        AChild.FindPointsInBlock(Block, List);
      end;

      if Block.YMax >= ymidLocal then
      begin
        //North Children are candidates.
        AChild := Children[North, East];
        AChild.FindPointsInBlock(Block, List);
      end;
    end
  end;
end;

procedure TRbwQuadTree.FindPointsInBlock(const Block: T2DBlock;
  var Points: TQuadPointInRegionArray);
var
  List : TList;
  PointIndex, DataIndex : integer;
  APoint : TQPoint;
  PointAddress : PQPoint;
begin
  List := TList.Create;
  try
    List.Capacity := Count;
    FQTreeNode.FindPointsInBlock(Block, List);
    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      PointAddress := List[PointIndex];
      APoint := PointAddress^;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;
  finally
    List.Free;
  end;
end;



procedure TRbwQuadTree.FindPointsInCircle(const CenterX, CenterY,
  Radius: double; var Points: TQuadPointInRegionArray);
var
  List : TList;
  PointIndex, DataIndex : integer;
  APoint : TQPoint;
  PointAddress : PQPoint;
begin
  List := TList.Create;
  try
    FQTreeNode.FindPointsInCircle(CenterX, CenterY, Radius, Sqr(Radius), List);
    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      PointAddress := List[PointIndex];
      APoint := PointAddress^;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;
  finally
    List.Free;
  end;
end;

procedure TRbwQuadTree.FindNearestPoints(const CenterX, CenterY: double;
  const Count: Integer; var Points: TQuadPointArray);
var
  List : TObjectList;
  PointIndex, DataIndex : integer;
  APoint : TQPoint;
  Node : TSelectNode;
  Radius  : double;
begin
  List := TObjectList.Create;
  try
    FQTreeNode.FindNearestPoints(CenterX, CenterY, Count, List);

    List.Sort(SelectNodeCompare);
    if List.Count > Count then
    begin
      Node := List[Count-1] as TSelectNode;
      Radius := Node.Distance;

      for PointIndex := Count to List.Count - 1 do
      begin
        Node := List[PointIndex] as TSelectNode;
        if Node.Distance > Radius then
        begin
          List.Count := PointIndex;
          break;
        end;
      end;
    end;


    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      Node := List[PointIndex] as TSelectNode;
      APoint := Node.Point;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      Points[PointIndex].Distance := Node.Distance;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;

  finally
    List.Free;
  end;
end;

procedure TRbwQuadTree.SetMaxPoints(const Value: integer);
begin
  if Count = 0 then
  begin
    if Value <= 0 then
    begin
      raise EQTreeError.Create('The maximum number of points in a Quadtree '
      + 'must be greater than zero.');
    end;
    FMaxPoints := Value;
    SetLength(FQTreeNode.Pts, FMaxPoints);
  end
  else
  begin
    raise EQTreeError.Create('You must set the number of points before adding '
      + 'any points to a Quadtree.');
  end;

end;

procedure Register;
begin
  RegisterComponents('RBW', [TRbwQuadTree]);
end;

procedure TQtreeNode.Clear;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  Index : integer;
begin
  if Children[North, East] = nil then
  begin
    for Index := 0 to NumPts-1 do
    begin
      Pts[Index].Data.Free;
      Pts[Index].Data := nil;
    end;
    FNumPts := 0;
  end
  else
  begin
    for ns := North to South do
    begin
      for ew := East to West do
      begin
        Children[ns,ew].Clear;
        Children[ns,ew].Free;
        Children[ns,ew] := nil;
      end;
    end;
  end;
end;

procedure TRbwQuadTree.Clear;
begin
  FQTreeNode.Clear;
  FQTreeNode.FNumPts := 0;
  SetLength(FQTreeNode.Pts, MaxPoints);
end;

end.
